from seeing.sympyHelpers import *
from seeing.formulary import *
from seeing.integrator import *


def createPropagationFormulary(
        cartesian=False,
        infinite_domain=False,
        circleLimits=True):

    approximations = [
    "Rayleigh-Sommerfeld",
    "Approximate Rayleigh-Sommerfeld",
    "Near Fresnel",
    "Far Fresnel",
    "Fraunhofer"]

    x0, y0, x1, y1, theta0, theta1 = sp.symbols('x_0, y_0 x_1, y_1 theta_0 theta_1', real=True)
    k, ll, a, z1, focal, rho, r1, r0, u, FN = sp.symbols('k lambda a z_1 f rho r_1 r_0 u FN', real=True, positive=True)
    ez0, ez0r = sp.symbols('E_0 E_0r')
    ez1 = sp.Function('E_1')(x1, y1, z1)
    ez1r = sp.Function('E_1r')(r1, theta1, z1)

    k = 2 * sp.S.Pi / ll
    
    def xyCircle():
        rrr = sp.S(1) - (x0 / sp.S(a))**sp.S(2) - (y0 / sp.S(a))**sp.S(2)
        return (rrr / sp.Abs(rrr) + 1) / 2


    def xyLens():
        ffn = FN * a * 2
        kl = sp.S(2) * sp.pi / ll
        rr1 = x0**sp.S(2) + y0**sp.S(2)
        return sp.exp(- sp.I * (kl * rr1 / (2 * ffn)))


    def rLens():
        ffn = FN * a * 2
        kl = sp.S(2) * sp.pi / sp.S(ll)
        return sp.exp(-sp.I * (kl / (2 * ffn)) * (r0**2))


    def getPropagationMethod(
            approximation="Rayleigh-Sommerfeld",
            infinite_domain=False,
            cartesian=False,
            circleLimits=True):

        _cartesian_circle_limits = (x0, -a, a), (y0, -sp.sqrt(sp.Abs(a * a - x0 * x0)), sp.sqrt(sp.Abs(a * a - x0 * x0)))
        _cartesian_limits = (x0, -a, a), (y0, -a, a)
        _radial_limits = (theta0, 0, sp.S(2) * sp.S.Pi), (r0, 0, a)
        _cartesian_limits_inf = (x0, -sp.oo, sp.oo), (y0, -sp.oo, sp.oo)
        _radial_limits_inf = [(r0, 0, sp.oo)]

        if not cartesian:
            if infinite_domain:
                _limits = _radial_limits_inf
            else:
                _limits = _radial_limits
        else:
            if infinite_domain:
                _limits = _cartesian_limits_inf
            else:
                if circleLimits:
                    _limits = _cartesian_circle_limits
                else:
                    _limits = _cartesian_limits

        iexpr23 = (k * z1 / (sp.I * 2 * sp.S.Pi)) * (sp.exp(sp.I * k * rho) / (rho**2))
        iexpr01 = iexpr23 * (1 - 1 / (sp.I * k * rho))

        common_fresnel_term0 = (sp.exp(sp.I * k * z1) / (z1))
        common_fresnel_term1 = (k / (2 * sp.S.Pi * sp.I))

#       common_fresnel_term = ( k * sp.exp(sp.I*k*z1)/(sp.I*2*sp.S.Pi*z1) )
        common_fresnel_term = common_fresnel_term0 * common_fresnel_term1

        iexpr4 = (common_fresnel_term) * sp.exp(sp.I * k * (r1**2) / (2 * z1)) * \
            sp.exp(sp.I * k * (r0**2) / (2 * z1)) * sp.besselj(0, k * r0 * r1 / z1)
        iexpr5 = common_fresnel_term * \
            sp.exp(sp.I * k * ((x1 - x0)**2 + (y1 - y0)**2) / (2 * z1))
        iexpr6 = common_fresnel_term * \
            sp.besselj(0, k * r0 * r1 / z1) * sp.exp(sp.I * k * (r0**2) / (2 * z1))
        iexpr7 = common_fresnel_term * \
            sp.exp(sp.I * k * (x0 * x1 + y0 * y1) / z1) * sp.exp(sp.I * k * (x0**2 + y0**2) / (2 * z1))
        iexpr8 = sp.exp(-sp.I * k * r0 * r1 * sp.sin(theta1 - theta0 + sp.S.Pi / 2) / z1) * k
        iexpr9 = sp.exp(-sp.I * k * (x0 * x1 + y0 * y1) / z1) * k * 2 * sp.S.Pi

        if approximation == "Rayleigh-Sommerfeld":
            if not cartesian:
                _integrand = iexpr01.subs(rho, sp.sqrt(
                    (x1 - x0)**2 + (y1 - y0)**2 + z1**2))
                dsub = {
                    x1: r1 * sp.cos(theta1),
                    x0: r0 * sp.cos(theta0),
                    y0: r0 * sp.sin(theta0),
                    y1: r1 * sp.sin(theta1)}
                _integrand = _integrand.subs(dsub)
            else:
                _integrand = iexpr01.subs(rho, sp.sqrt(
                    (x1 - x0)**2 + (y1 - y0)**2 + z1**2))
        elif approximation == "Approximate Rayleigh-Sommerfeld":
            if not cartesian:
                _integrand = iexpr23.subs(rho, sp.sqrt(
                    (x1 - x0)**2 + (y1 - y0)**2 + z1**2))
                dsub = {
                    x1: r1 * sp.cos(theta1),
                    x0: r0 * sp.cos(theta0),
                    y0: r0 * sp.sin(theta0),
                    y1: r1 * sp.sin(theta1)}
                _integrand = _integrand.subs(dsub)
            else:
                _integrand = iexpr23.subs(rho, sp.sqrt(
                    (x1 - x0)**2 + (y1 - y0)**2 + z1**2))
        elif approximation == "Near Fresnel":
            if not cartesian:
                _integrand = iexpr4
            else:
                _integrand = iexpr5
        elif approximation == "Far Fresnel":
            if not cartesian:
                _integrand = iexpr6
            else:
                _integrand = iexpr7
        elif approximation == "Fraunhofer":
            if not cartesian:
                _integrand = iexpr8
            else:
                _integrand = iexpr9

        if not cartesian:
            _integrand = 2 * sp.S.Pi * r0 * ez0r * _integrand
        else:
            _integrand *= ez0

        _name = approximation + "Integral"
        if cartesian:
            _name += ', ' + "cartesian coords"
        else:
            _name += ', ' + "cilyndrical coords"
        if infinite_domain:
            _name += ', ' + "infinite domain"
        else:
            _name += ', ' + "finite domain"

        return (_name, _integrand, _limits)
    
    
    prop_f = Formulary()
    for appr in approximations:
        _name, _integrand, _limits = getPropagationMethod(
            appr, infinite_domain, cartesian, circleLimits)
        _integral = sp.Integral(_integrand, *_limits)
        if cartesian:
            prop_f.addFormula(appr, sp.Eq(ez1, _integral))
            prop_f.addFormula(appr+" Arg", _integral.function)
        else:
            prop_f.addFormula(appr, sp.Eq(ez1r, _integral))

    if cartesian:
        prop_f.addFormula('xyCircle', xyCircle() )
        prop_f.addFormula('xyLens', xyLens() )
    else:
        prop_f.addFormula('rLens', rLens() )
    
    return prop_f


'''
class propagationStep(object):
    def __init__(method, lll, aa, subdiv_points):
        self.lll = lll
        #lll = 500e-9
        self.aa = aa
        self.subdiv_points = subdiv_points
        #aa =  0.05
        #FN = 30
        #dd = 2*aa
        #focal_dist = dd*FN
        self.method = method

    def setsp.Input(Ez0):
        self.Ez0 = Ez0

    def propagate(distance, outputRadius):
        subdiv_points = self.subdiv_points
        paramAndRange2 = [('x1', -outputRadius, outputRadius, subdiv_points, 'linear'), ('y1', -outputRadius, outputRadius, subdiv_points, 'linear')]
        # dictionary for Fraunhofer method, at the focal plane of a lens
        subsDictC = {ez0: self.Ez0, ll: self.lll, a:self.aa}
        ff = propMethodsCartesian
        sDict = subsDictC
        pr = paramAndRange2
        (lh, rh, eeq) = ff.getFormula(self.method)
        sDict[z1] = distance
        rh = rh.subs(sDict)
        fplot1 = msp.It.sp.IntegralEval(lh, rh, pr, [(subdiv_points, 'linear'), (subdiv_points, 'linear')])
        return fplot1
'''



import numpy as np
import sympy as sp

from seeing.sympyHelpers import *
from seeing.formulary import *
from seeing.integrator import *


def noll_to_zern(jj):
    if (jj == 0):
        raise ValueError("Noll indices start at 1, 0 is invalid.")
    nn = 0
    jj1 = jj - 1
    while (jj1 > nn):
        nn += 1
        jj1 -= nn
    mm = (-1)**jj * ((nn % 2) + 2 * int((jj1 + ((nn + 1) % 2)) / 2.0))
    return nn, mm


def order_to_max_noll(nn):
    return (nn + 2) * (nn + 1) // 2


def NollIndex(nn, mm):
    n4 = nn % 4
    if (mm > 0 and (n4 == 0 or n4 == 1)) or (mm < 0 and (n4 == 2 or n4 == 3)):
        k = 0
    else:
        k = 1
    return nn * (nn + 1) // 2 + abs(mm) + k


def kronDelta(ii, jj):
    if ii == jj:
        return 1
    else:
        return 0


def emValue(mm):
    if mm == 0:
        return 2
    else:
        return 1


def createZernikeFormulary():
    
    r, rho = sp.symbols('r rho', positive=True)
    aa = sp.symbols('a')
    l, j, n, m, max_order = sp.symbols('l j n m n_max', integer=True)
    vlj, theta, f, x0, y0 = sp.symbols('v_lj theta f x0, y0', real=True)
    dotInt = sp.symbols('I_v')
    Cnm = sp.symbols('c_n^m')

    Tnm = sp.Function("T_n^m")(theta)
    Rnm = sp.Function("R_n^m")(rho)
    Znm = sp.Function("Z_n^m")(rho, theta)
    Ynm = sp.Function("Y_n^m")(rho, theta)
    ZnmCart = sp.Function("Z_n^m")(x0, y0)
    VnmFunc = sp.Function("V_n^m")(r)
    ZnmD = sp.Function("Z_n^m_d")(r, theta)
    YnmD = sp.Function("Y_n^m_d")(r, theta)
    f1 = sp.Function("f_1")(rho, theta)
    f2 = sp.Function("f_2")(rho, theta)

    # Real Zernike Polynomials
    def cCoeff():
        _rhs = sp.Piecewise((sp.sqrt(n + 1), m == 0),
                            (sp.sqrt(2 * (n + 1)), m != 0))
        _lhs = Cnm
        return sp.Eq(_lhs, _rhs)


    def tangetialFunc():
        _rhs = sp.Piecewise((-sp.sin(m * theta), m < 0),
                            (sp.cos(m * theta), m >= 0))
        _lhs = Tnm
        return sp.Eq(_lhs, _rhs)


    def complexTangetialFunc():
        _rhs = sp.exp(sp.I * m * theta)
        _lhs = Tnm
        return sp.Eq(_lhs, _rhs)


    def radialFunc():
        ma = sp.Abs(m)
        p = (n - ma) / 2
        _rhs = ((-1)**p) * (rho**ma) * sp.jacobi(p, ma, 0, 1 - 2 * rho**2)
        _lhs = Rnm
        return sp.Eq(_lhs, _rhs)


    def realZernike():
        _rhs = radialFunc().rhs * tangetialFunc().rhs
        _lhs = Znm
        return sp.Eq(_lhs, _rhs)


    def realZernikeCartesian():
        _rhs = realZernike().rhs.subs(rho,sp.sqrt(x0**2+y0**2)/aa).subs(theta,sp.atan2(x0,y0))
        _lhs = ZnmCart
        return sp.Eq(_lhs, _rhs)


    def realZernikeNormalized():
        _rhs = cCoeff().rhs * realZernike().rhs
        _lhs = Znm
        return sp.Eq(_lhs, _rhs)

    # Complex Zernike Polynomials

    def complexZernike():
        _rhs = sp.sqrt(2) * radialFunc().rhs * complexTangetialFunc().rhs / 2
        _lhs = Ynm
        return sp.Eq(_lhs, _rhs)


    def complexZernikeNormalized():
        _rhs = sp.sqrt(n + 1) * radialFunc().rhs * complexTangetialFunc().rhs
        _lhs = Ynm
        return sp.Eq(_lhs, _rhs)

    # with this wiegths and this transformation scheme, (realZernike,
    # complexZernike) and (realZernikeNormalized, complexZernikeNormalized)
    # are coerent


#    def realZFromComplexZ(zcf, nn, mm):
#        zc1 = zcf(n, m).subs({m: mm, n: nn})
#        zc2 = zcf(n, m).subs({m: -mm, n: nn})
#        if mm >= 0:
#            _rhs = (zc1 + zc2) / np.sqrt(2)
#        elif mm < 0:
#            _rhs = 1j * (zc1 - zc2) / np.sqrt(2)    
#        _lhs = Znm
#        return sp.Eq(_lhs, _rhs)


    def _vlj():
        p = (n - m) / 2
        q = (n + m) / 2
        _rhs = (-1)**p * (m + l + j * 2) * sp.binomial(m + j + l - 1,
                                                       l - 1) * sp.binomial(j + l - 1,
                                                                            l - 1) * sp.binomial(l - 1,
                                                                                                 p - j) / sp.binomial(q + l + j,
                                                                                                                      l)
        _lhs = vlj
        return sp.Eq(_lhs, _rhs)


    def Vnm():
        ma = sp.Abs(m)
        p = (n - ma) / 2
        q = (n + ma) / 2
        _rhs = sp.exp(sp.I * f) * sp.Sum((-2 * sp.I * f)**(l - 1) * sp.Sum((-1)**p * (ma + l + j * 2) * sp.binomial(ma + j + l - 1,
                                                                                                              l - 1) * sp.binomial(j + l - 1,
                                                                                                                                   l - 1) * sp.binomial(l - 1,
                                                                                                                                                        p - j) / sp.binomial(q + l + j,
                                                                                                                                                                             l) * (sp.besselj(ma + l + j * 2,
                                                                                                                                                                                              2 * sp.pi * r) / (l * (2 * sp.pi * r)**l)),
                                                                     (j,
                                                                      0,
                                                                      p)),
                                      (l,
                                       1,
                                       max_order))
        _lhs = VnmFunc
        return sp.Eq(_lhs, _rhs)


    def diffractedZernikeAtFocus():
        #    return ((-1)**((n+m)/2) * sp.besselj( n+1, 2*sp.pi*r) / (2*sp.pi*r) ) * tangetialFunc(m)
        # return 2*sp.pi*(-1**(n+1)) * ( sp.besselj( n+1, 2*sp.pi*r) / (2*sp.pi*r)
        # ) * sp.Piecewise( (-sp.sin(m*theta), m<0), (sp.cos(m*theta), m>=0) )
        _rhs = 2 * sp.pi * (sp.besselj(n + 1, 2 * sp.pi * r) / (2 * sp.pi * r)) * \
            sp.Piecewise((-sp.sin(m * theta), m < 0), (sp.cos(m * theta), m >= 0))
        _lhs = ZnmD
        return sp.Eq(_lhs, _rhs)


    def diffractedComplexZernikeAtFocus():
        # return 2*sp.pi* (-1**(n+1)) * ( sp.besselj( n+1, 2*sp.pi*r) /
        # (2*sp.pi*r) ) * complexTangetialFunc(m) / sp.sqrt(2)
        _rhs = 2 * sp.pi * (sp.besselj(n + 1, 2 * sp.pi * r) /
                            (2 * sp.pi * r)) * complexTangetialFunc().rhs / sp.sqrt(2)
        _lhs = YnmD
        return sp.Eq(_lhs, _rhs)


    def diffractedZernike():
        _rhs = 2 * Vnm().rhs * tangetialFunc().rhs
        _lhs = ZnmD
        return sp.Eq(_lhs, _rhs)


    def diffractedComplexZernike():
        _rhs = 2 * Vnm().rhs * complexTangetialFunc().rhs / sp.sqrt(2)
        _lhs = YnmD
        return sp.Eq(_lhs, _rhs)


    def checkOrthoPair():
        f2c = sp.conjugate(f2)
        _rhs = sp.Integral(f1 * f2 * rho, (rho, 0, 1), (theta, 0, 2 * sp.pi))
        _lhs = dotInt
        return sp.Eq(_lhs, _rhs)

    
    def circleDotProduct():
        f2c = sp.conjugate(f2)
        itR = sp.re(f1 * f2c * rho) / sp.S.Pi
        itI = sp.im(f1 * f2c * rho) / sp.S.Pi
        i1 = sp.N(sp.integrate(itR, (theta, 0, 2 * sp.pi), (rho, 0, 1)))
        i2 = sp.N(sp.integrate(itI, (theta, 0, 2 * sp.pi), (rho, 0, 1)))
        _rhs = i1 + sp.I * i2
        _lhs = dotInt
        return sp.Eq(_lhs, _rhs)

    _zernikeFormulas = Formulary("Zernike",                                                                          
                                   ['cCoeff',
                                    'tangetialFunc',
                                    'complexTangetialFunc',
                                    'radialFunc',
                                    'realZernike',
                                    'realZernikeCartesian',
                                    'realZernikeNormalized',
                                    'complexZernike',
                                    'complexZernikeNormalized',
                                    #'realZFromComplexZ',
                                    '_vlj',
                                    'Vnm',
                                    'diffractedZernikeAtFocus',
                                    'diffractedComplexZernikeAtFocus',   
                                    'diffractedZernike',
                                    'diffractedComplexZernike',
                                    'checkOrthoPair',
                                    #'checkZernikeOrthoPair',
                                    'circleDotProduct',
#                                    'zernikeAnalysysReal',
#                                    'zernikeSynthesysReal',
#                                    'zernikeAnalysysComplex',
#                                    'zernikeSynthesysComplex',
                                    #'capital_v_vector',
#                                    'createZernikeFormulary'
                                   ],
                                  [  cCoeff(),
                                     tangetialFunc(),
                                     complexTangetialFunc(),
                                     radialFunc(),
                                     realZernike(),
                                     realZernikeCartesian(),
                                     realZernikeNormalized(),
                                     complexZernike(),
                                     complexZernikeNormalized(),
                                     #realZFromComplexZ(),
                                     _vlj(),
                                     Vnm(),
                                     diffractedZernikeAtFocus(),
                                     diffractedComplexZernikeAtFocus(),
                                     diffractedZernike(),
                                     diffractedComplexZernike(),
                                     checkOrthoPair(),
                                     #checkZernikeOrthoPair(),
                                     circleDotProduct(),
#                                     zernikeAnalysysReal(),
#                                     zernikeSynthesysReal(),
#                                     zernikeAnalysysComplex(),
#                                     zernikeSynthesysComplex(),
                                     #capital_v_vector(),
#                                     createZernikeFormularyOld()
                                  ])

    return _zernikeFormulas


def cov_expr_jk(expr, jj_value, kk_value):
    nj_value, mj_value = noll_to_zern(jj_value)
    nk_value, mk_value = noll_to_zern(kk_value)
    rexpr = subsParamsByName(expr, {'j': jj_value, 'k': kk_value, 'n_j': nj_value, 'm_j': abs(mj_value), 'n_k': nk_value, 'm_k': abs(mk_value)})
    return rexpr


def getZernikeDomain(nn):
    x1 = np.linspace(-1.0, 1.0, nn)
    y1 = np.linspace(-1.0, 1.0, nn)
    X1, Y1 = np.meshgrid(x1, y1)
    rr = np.sqrt(X1**2 + Y1**2)
    rr[np.where(rr > 1.0)] = np.nan
    return rr, np.arctan2(Y1, X1)
    

def evaluateZernike(zerninke_mode_expression, sampling_points):
    zerninke_mode_lambda = sp.lambdify(
        [rho, theta], zerninke_mode_expression, 'numpy')
#    r1 = 1.0 - np.geomspace(1.0/sampling_points, 1.0, sampling_points, endpoint=True)
    r1 = np.power(np.linspace(0.0, 1.0, sampling_points), 1.0 / 2.0)
    theta1 = np.linspace(0, 2 * np.pi, sampling_points)
    r1, theta1 = np.meshgrid(r1, theta1)
    X1 = r1 * np.sin(theta1)
    Y1 = r1 * np.cos(theta1)
    Z1 = np.asarray(zerninke_mode_lambda(r1, theta1))
    return X1, Y1, Z1

'''
    def checkZernikeOrthoPair(n1, m1, n2, m2):
        f1 = realZernikeNormalized(n1, m1)
        f2 = realZernikeNormalized(n2, m2)
        _rhs = checkOrthoPair(f1, f2)
        _lhs = Cnm
        return sp.Eq(_lhs, _rhs)

    # inputFunction defined over the unit disc, in ro, theta
    def zernikeAnalysysReal(inputFunction, max_noll):
        result = [0] * max_noll
        for ni in range(max_noll):
            nn, mm = noll_to_zern(ni + 1)
            z_ni = realZernikeNormalized(n, m).subs({m: int(mm), n: int(nn)})
            result[ni] = circleDotProduct(inputFunction, z_ni).rhs
            print(ni + 1, nn, mm, result[ni] / emValue(mm))
        _rhs = result
        _lhs = Cnm
        return sp.Eq(_lhs, _rhs)

    def zernikeSynthesysReal(decomposition):
        result = sp.S(0)
        for ni, coefficient in enumerate(decomposition):
            nn, mm = noll_to_zern(ni + 1)
            z_ni = realZernikeNormalized(n, m).subs({m: int(mm), n: int(nn)})
            result += coefficient * z_ni
            print(ni + 1, nn, mm, coefficient)
        _rhs = result
        _lhs = Cnm
        return sp.Eq(_lhs, _rhs)


    def zernikeAnalysysComplex(inputFunction, max_noll):
        result = [0] * max_noll
        for ni in range(max_noll):
            nn, mm = noll_to_zern(ni + 1)
            z_ni = complexZernikeNormalized(n, m).subs({m: mm, n: nn})
            result[ni] = circleDotProduct(inputFunction, z_ni)
            print(ni + 1, nn, mm, result[ni])
        _rhs = result
        _lhs = Cnm
        return sp.Eq(_lhs, _rhs)


    def zernikeSynthesysComplex(decomposition):
        result = sp.S(0)
        for ni, coefficient in enumerate(decomposition):
            nn, mm = noll_to_zern(ni + 1)
            z_ni = complexZernikeNormalized(n, m).subs({m: mm, n: nn})
            result += coefficient * z_ni
            print(ni + 1, nn, -mm, coefficient)
        _rhs = result
        _lhs = Cnm
        return sp.Eq(_lhs, _rhs)


    def capital_v_vector(max_noll, f):
        result = [0] * max_noll
        for ni in range(max_noll):
            nn, mm = noll_to_zern(ni + 1)
            z_ni = CVVZernike(nn, mm, rho, theta)
            result[ni] = z_ni
        _rhs = result
        _lhs = Cnm
        return sp.Eq(_lhs, _rhs)
        
        
    def createZernikeFormularyOld(lastMode):
        zf = Formulary()
        for i in range(2, lastMode + 1):
            idx = noll_to_zern(i)
            idstr = str(idx[0]) + str(idx[1])
            zzz = sp.symbols('Z_' + idstr)
            az = realZernike(n, m).subs({m: int(idx[1]), n: int(idx[0])})
            zname = 'Z' + idstr
            zf.addFormula(zname, (zzz, az, sp.Eq(zzz, az)))
        return zf
'''




from scipy.special import *
import scipy
import time
import random
import sympy as sp

from seeing.sympyHelpers import *
from seeing.formulary import *
from seeing.integrator import *

def createTurbolenceFormulary():
    #R_q = sp.Function("\U0000211B_q")(r)
    V0, L0, ni0, Re, l0, r0, r, c = sp.symbols(r'V_0 L_0 \nu_0 Re l_0 r_0 r c', positive=True)
    h, f, k, k0, km, gamma = sp.symbols(r'h f k k_0 k_m \gamma', positive=True)
    dx, dt, v = sp.symbols("\u03B4x \u03B4t v", positive=True)
    cn2 = sp.Function("C_N")(h)
    dPhi = sp.Function("D_phi")(dx, dt)
    dPhi_r = sp.Function("D_phi")(r)
    cPhi_r = sp.Function("C_phi")(r)
    pPhi = sp.Function("P_phi")(k)
    wPhi = sp.Function("W_phi")(f)
    
    def ReynoldsNumber():
        _lhs = Re
        _rhs = V0 * L0 / ni0
        return sp.Eq(_lhs, _rhs)


    def innerScale():
        _lhs = l0
        _rhs = L0 / Re ** (sp.S(3) / sp.S(4))
        return sp.Eq(_lhs, _rhs)


    def FriedParameter():
        _lhs = r0
        with sp.evaluate(False):
            _rhs = ((0.423 * k**2 / sp.cos(gamma)) *
                    sp.Integral(cn2, (h, 0, sp.oo))) ** (-sp.S(3) / sp.S(5))
        return sp.core.relational.Eq(_lhs, _rhs)


    def phaseStructureFunctionOrig():
        _lhs = dPhi
        _rhs = 6.88 * (sp.Abs(dx - v * dt) / r0) ** (sp.S(5) / sp.S(3))
        return sp.core.relational.Eq(_lhs, _rhs)


    def phaseStructureFunctionOrig_r():
        _lhs = dPhi_r
        with sp.evaluate(False):
            _rhs = 6.88 * (r / r0)**(sp.S(5) / sp.S(3))
        return sp.core.relational.Eq(_lhs, _rhs)


    def phaseStructureFunction():
        _lhs = dPhi_r
        expr0 = c * (L0 / r0)**(sp.S(5) / sp.S(3)) * \
            sp.gamma(sp.S(5) / sp.S(6)) / (sp.S(2)**(sp.S(1) / sp.S(6)))
        frac1 = sp.S(5) / sp.S(3)
        frac2 = sp.S(5) / sp.S(6)
        frac3 = sp.S(1) / sp.S(6)
        frac4 = sp.S(11) / sp.S(6)
        frac5 = sp.S(8) / sp.S(3)
        frac6 = sp.S(24) / sp.S(5)
        frac7 = sp.S(6) / sp.S(5)
        c_expr = (sp.S(2)**frac3 * sp.gamma(frac4) / sp.pi**frac5) * \
            (frac6 * sp.gamma(frac7)) ** frac2
        _rhs = D_expr = c_expr * (L0 / r0)**frac1 * (sp.gamma(frac2) / sp.S(2)**frac3 - (
            2 * sp.pi * r / L0)**frac2 * sp.besselk(frac2, 2 * sp.pi * r / L0))
        return sp.core.relational.Eq(_lhs, _rhs)


    def phaseVariance():
        _lhs = cPhi_r
        frac1 = sp.S(5) / sp.S(3)
        frac2 = sp.S(5) / sp.S(6)
        frac3 = sp.S(1) / sp.S(6)
        frac4 = sp.S(11) / sp.S(6)
        frac5 = sp.S(8) / sp.S(3)
        frac6 = sp.S(24) / sp.S(5)
        frac7 = sp.S(6) / sp.S(5)
        c_expr = (sp.S(2)**frac3 * sp.gamma(frac4) / sp.pi**frac5) * \
            (frac6 * sp.gamma(frac7)) ** frac2
        _rhs = C_expr = (L0 / r0) ** frac1 * (c_expr / 2) * (2 *
                                                             sp.pi * r / L0)**frac2 * sp.besselk(frac2, 2 * sp.pi * r / L0)
        return sp.core.relational.Eq(_lhs, _rhs)


    #P_expr = 0.00058 * r0**(-sp.S(5)/sp.S(3)) * (f**2 + (1/L0**2)) ** ( -sp.S(11) / sp.S(6) )
    def phaseSpatialPowerSpectrumKolmogorov():
        _lhs = pPhi
        with sp.evaluate(False):
            _rhs = 0.0229 * r0**(-sp.S(5) / sp.S(3)) * k ** (-sp.S(11) / sp.S(3))

        return sp.core.relational.Eq(_lhs, _rhs)


    def phaseSpatialPowerSpectrumVonKarmanO():
        _lhs = pPhi
        k0 = 2 * sp.pi / L0
        with sp.evaluate(False):
            _rhs = 0.0229 * r0**(-sp.S(5) / sp.S(3)) * \
                (k**2 + k0**2) ** (-sp.S(11) / sp.S(6))

        return sp.core.relational.Eq(_lhs, _rhs)


    def phaseSpatialPowerSpectrumVonKarmanO_f():
        _lhs = wPhi
        f0 = 1 / L0
        with sp.evaluate(False):
            _rhs = 0.0229 * r0**(-sp.S(5) / sp.S(3)) * \
                (f**2 + f0**2) ** (-sp.S(11) / sp.S(6))

        return sp.core.relational.Eq(_lhs, _rhs)


    def phaseSpatialPowerSpectrumVonKarman():
        # Von-Karman outer/inner
        _lhs = pPhi
        k0 = 2 * sp.pi / L0
        km = 5.92 / l0
        with sp.evaluate(False):
            _rhs = 0.0229 * r0**(-sp.S(5) / sp.S(3)) * (k**2 + k0 **
                                                        2) ** (-sp.S(11) / sp.S(6)) * sp.exp(-(k * l0 / 5.92)**2)

        return sp.core.relational.Eq(_lhs, _rhs)
    
    '''
    def r0_to_seeing(r0, lamda=500.E-9):
        return (0.98*lamda/r0)*180.*3600./np.pi

    def r0_to_cn2(r0, lamda=500.E-9):
        cn2 = r0**(-5./3.)/(0.423*(2*np.pi/lamda)**2)
        return cn2

    def cn2_to_r0(cn2, lamda=500.E-9):
        r0=(0.423*(2*np.pi/lamda)**2*cn2)**(-3./5.)
        return r0

    def cn2_to_seeing(cn2, lamda=500.E-9):
        r0 = cn2_to_r0(cn2,lamda)
        seeing = r0_to_seeing(r0,lamda)
        return seeing

    def seeing_to_cn2(seeing, lamda=500.E-9):
        r0 = seeing_to_r0(seeing,lamda)
        cn2 = r0_to_cn2(r0,lamda)
        return cn2

    def coherenceTime(cn2, v, lamda=500.E-9):
        Jv = (cn2*(v**(5./3.))).sum()
        tau0 = float((Jv**(-3./5.))*0.057*lamda**(6./5.))
        return tau0

    def isoplanaticAngle(cn2, h, lamda=500.E-9):
        Jh = (cn2*(h**(5./3.))).sum()
        iso = float(0.057*lamda**(6./5.)*Jh**(-3./5.)*180.*3600./np.pi)
        return iso

    def r0_from_slopes(slopes, wavelength, subapDiam):
        slopeVar = slopes.var(axis=(-1))
        r0 = ((0.162 * (wavelength ** 2) * subapDiam ** (-1. / 3)) / slopeVar) ** (3. / 5)
        r0 = float(r0.mean())
        return r0

    def slope_variance_from_r0(r0, wavelength, subapDiam):
        slope_var = 0.162 * (wavelength ** 2) * r0 ** (-5. / 3) * subapDiam ** (-1. / 3)
        return slope_var
    
'''
    
    
    _turbolenceFormulas = Formulary("Turbolence",
                                   ['ReynoldsNumber',
                                    'innerScale',
                                    'FriedParameter',
                                    'phaseStructureFunctionOrig',
                                    'phaseStructureFunctionOrig_r',
                                    'phaseStructureFunction',
                                    'phaseVariance',
                                    'phaseSpatialPowerSpectrumKolmogorov',
                                    'phaseSpatialPowerSpectrumVonKarmanO',
                                    'phaseSpatialPowerSpectrumVonKarmanO_f',
                                    'phaseSpatialPowerSpectrumVonKarman'],
                                   [ReynoldsNumber(),
                                       innerScale(),
                                       FriedParameter(),
                                       phaseStructureFunctionOrig(),
                                       phaseStructureFunctionOrig_r(),
                                       phaseStructureFunction(),
                                       phaseVariance(),
                                       phaseSpatialPowerSpectrumKolmogorov(),
                                       phaseSpatialPowerSpectrumVonKarmanO(),
                                       phaseSpatialPowerSpectrumVonKarmanO_f(),
                                       phaseSpatialPowerSpectrumVonKarman()])


    return _turbolenceFormulas


pplib = {
    'factorial': factorial,
    'binomial': binom,
    'besselj': jv,
    'besselk': kv,
    'besseli': iv,
    'bessely': yv,
    'erf': erf,
    'gamma': gamma}

scipyext = [pplib, "scipy"]


def ft_ift2(G, delta_f):
    N = G.shape[0]
    g = np.fft.fftshift(np.fft.ifft2(np.fft.fftshift(G))) * (N * delta_f)**2
    return g


def ft_PSD_phi(rrr0, N, frq_range, L0, l0, method='VonKarman'):
    del_f = frq_range / N
    fx = np.arange(-N / 2., N / 2.) * del_f
    (fx, fy) = np.meshgrid(fx, fx)
    f = np.sqrt(fx**2 + fy**2)
    if method == 'VonKarman':
        expr = turbolenceFormulas['phaseSpatialPowerSpectrumVonKarman'].rhs  
        (_, PSD_phi) = evaluateFormula( expr, {'r_0': rrr0, 'L_0': L0, 'l_0': l0}, ['k'] , [2 * np.pi * f], cpulib)
    else:
        expr = turbolenceFormulas['phaseSpatialPowerSpectrumKolmogorov'].rhs 
        (_, PSD_phi) = evaluateFormula( expr, {'r_0': rrr0}, ['k'],  [2 * np.pi * f], cpulib)
        
    PSD_phi[int(N / 2), int(N / 2)] = 0
    return PSD_phi, del_f


def ft_phase_screen(rrr0, N, delta, L0, l0, method='VonKarman', seed=None):
    delta = float(delta)
    R = random.SystemRandom(time.time())
    if seed is None:
        seed = int(R.random() * 100000)
    np.random.seed(seed)
    frq_range = 1.0 / delta
    del_f = frq_range / N
    PSD_phi = ft_PSD_phi(rrr0, N, frq_range, L0, l0, method)
    cn = (
        (np.random.normal(
            size=(
                N,
                N)) +
         1j *
         np.random.normal(
            size=(
                N,
                N))) *
        np.sqrt(PSD_phi[0]) *
        del_f)
    phs = ft_ift2(cn, 1).real
    return phs
